@@ -6,12 +6,12 @@ import (
 	"fmt"
 	"time"
 
-	"github.com/lestrrat-go/jwx/v2/jwa"
-	"github.com/lestrrat-go/jwx/v2/jwk"
-	"github.com/lestrrat-go/jwx/v2/jwt"
+	"github.com/lestrrat-go/jwx/v3/jwa"
+	"github.com/lestrrat-go/jwx/v3/jwk"
+	"github.com/lestrrat-go/jwx/v3/jwt"
 )
 
-func Example_jwt_serialize_jwe_and_jws() {
+func Example_jwt_serialize_jwe_jws() {
 	tok, err := jwt.NewBuilder().
 		Issuer(`github.com/lestrrat-go/jwx`).
 		IssuedAt(time.Unix(aLongLongTimeAgo, 0)).
@@ -27,21 +27,21 @@ func Example_jwt_serialize_jwe_and_jws() {
 		return
 	}
 
-	enckey, err := jwk.FromRaw(privkey.PublicKey)
+	enckey, err := jwk.Import(privkey.PublicKey)
 	if err != nil {
 		fmt.Printf("failed to create symmetric key: %s\n", err)
 		return
 	}
 
-	signkey, err := jwk.FromRaw([]byte(`abracadabra`))
+	signkey, err := jwk.Import([]byte(`abracadabra`))
 	if err != nil {
 		fmt.Printf("failed to create symmetric key: %s\n", err)
 		return
 	}
 
 	serialized, err := jwt.NewSerializer().
-		Encrypt(jwt.WithKey(jwa.RSA_OAEP, enckey)).
-		Sign(jwt.WithKey(jwa.HS256, signkey)).
+		Encrypt(jwt.WithKey(jwa.RSA_OAEP(), enckey)).
+		Sign(jwt.WithKey(jwa.HS256(), signkey)).
 		Serialize(tok)
 	if err != nil {
 		fmt.Printf("failed to encrypt and sign token: %s\n", err)
