@@ -1,27 +1,22 @@
 //go:generate ../tools/cmd/genjwk.sh
 
-// Package jwk implements JWK as described in https://tools.ietf.org/html/rfc7517
 package jwk
 
 import (
 	"bytes"
 	"crypto"
 	"crypto/ecdsa"
-	"crypto/ed25519"
-	"crypto/elliptic"
-	"crypto/rsa"
+	"crypto/x509"
 	"encoding/pem"
 	"errors"
 	"fmt"
 	"io"
 	"math/big"
+	"reflect"
+	"slices"
 
-	"github.com/lestrrat-go/jwx/v2/internal/base64"
-	"github.com/lestrrat-go/jwx/v2/internal/ecutil"
-	"github.com/lestrrat-go/jwx/v2/internal/json"
-	"github.com/lestrrat-go/jwx/v2/jwa"
-	"github.com/lestrrat-go/jwx/v2/jwk/internal/x509"
-	"github.com/lestrrat-go/jwx/v2/x25519"
+	"github.com/lestrrat-go/jwx/v3/internal/base64"
+	"github.com/lestrrat-go/jwx/v3/internal/json"
 )
 
 var registry = json.NewRegistry()
@@ -33,7 +28,24 @@ func bigIntToBytes(n *big.Int) ([]byte, error) {
 	return n.Bytes(), nil
 }
 
-// FromRaw creates a jwk.Key from the given key (RSA/ECDSA/symmetric keys).
+func init() {
+	if err := RegisterProbeField(reflect.StructField{
+		Name: "Kty",
+		Type: reflect.TypeFor[string](),
+		Tag:  `json:"kty"`,
+	}); err != nil {
+		panic(fmt.Errorf("failed to register mandatory probe for 'kty' field: %w", err))
+	}
+	if err := RegisterProbeField(reflect.StructField{
+		Name: "D",
+		Type: reflect.TypeFor[json.RawMessage](),
+		Tag:  `json:"d,omitempty"`,
+	}); err != nil {
+		panic(fmt.Errorf("failed to register mandatory probe for 'kty' field: %w", err))
+	}
+}
+
+// Import creates a jwk.Key from the given key (RSA/ECDSA/symmetric keys).
 //
 // The constructor auto-detects the type of key to be instantiated
 // based on the input type:
@@ -41,84 +53,21 @@ func bigIntToBytes(n *big.Int) ([]byte, error) {
 //   - "crypto/rsa".PrivateKey and "crypto/rsa".PublicKey creates an RSA based key
 //   - "crypto/ecdsa".PrivateKey and "crypto/ecdsa".PublicKey creates an EC based key
 //   - "crypto/ed25519".PrivateKey and "crypto/ed25519".PublicKey creates an OKP based key
+//   - "crypto/ecdh".PrivateKey and "crypto/ecdh".PublicKey creates an OKP based key
 //   - []byte creates a symmetric key
-func FromRaw(key interface{}) (Key, error) {
-	if key == nil {
-		return nil, fmt.Errorf(`jwk.FromRaw requires a non-nil key`)
+func Import(raw any) (Key, error) {
+	if raw == nil {
+		return nil, importerr(`a non-nil key is required`)
 	}
 
-	var ptr interface{}
-	switch v := key.(type) {
-	case rsa.PrivateKey:
-		ptr = &v
-	case rsa.PublicKey:
-		ptr = &v
-	case ecdsa.PrivateKey:
-		ptr = &v
-	case ecdsa.PublicKey:
-		ptr = &v
-	default:
-		ptr = v
+	muKeyImporters.RLock()
+	conv, ok := keyImporters[reflect.TypeOf(raw)]
+	muKeyImporters.RUnlock()
+	if !ok {
+		return nil, importerr(`failed to convert %T to jwk.Key: no converters were able to convert`, raw)
 	}
 
-	switch rawKey := ptr.(type) {
-	case *rsa.PrivateKey:
-		k := newRSAPrivateKey()
-		if err := k.FromRaw(rawKey); err != nil {
-			return nil, fmt.Errorf(`failed to initialize %T from %T: %w`, k, rawKey, err)
-		}
-		return k, nil
-	case *rsa.PublicKey:
-		k := newRSAPublicKey()
-		if err := k.FromRaw(rawKey); err != nil {
-			return nil, fmt.Errorf(`failed to initialize %T from %T: %w`, k, rawKey, err)
-		}
-		return k, nil
-	case *ecdsa.PrivateKey:
-		k := newECDSAPrivateKey()
-		if err := k.FromRaw(rawKey); err != nil {
-			return nil, fmt.Errorf(`failed to initialize %T from %T: %w`, k, rawKey, err)
-		}
-		return k, nil
-	case *ecdsa.PublicKey:
-		k := newECDSAPublicKey()
-		if err := k.FromRaw(rawKey); err != nil {
-			return nil, fmt.Errorf(`failed to initialize %T from %T: %w`, k, rawKey, err)
-		}
-		return k, nil
-	case ed25519.PrivateKey:
-		k := newOKPPrivateKey()
-		if err := k.FromRaw(rawKey); err != nil {
-			return nil, fmt.Errorf(`failed to initialize %T from %T: %w`, k, rawKey, err)
-		}
-		return k, nil
-	case ed25519.PublicKey:
-		k := newOKPPublicKey()
-		if err := k.FromRaw(rawKey); err != nil {
-			return nil, fmt.Errorf(`failed to initialize %T from %T: %w`, k, rawKey, err)
-		}
-		return k, nil
-	case x25519.PrivateKey:
-		k := newOKPPrivateKey()
-		if err := k.FromRaw(rawKey); err != nil {
-			return nil, fmt.Errorf(`failed to initialize %T from %T: %w`, k, rawKey, err)
-		}
-		return k, nil
-	case x25519.PublicKey:
-		k := newOKPPublicKey()
-		if err := k.FromRaw(rawKey); err != nil {
-			return nil, fmt.Errorf(`failed to initialize %T from %T: %w`, k, rawKey, err)
-		}
-		return k, nil
-	case []byte:
-		k := newSymmetricKey()
-		if err := k.FromRaw(rawKey); err != nil {
-			return nil, fmt.Errorf(`failed to initialize %T from %T: %w`, k, rawKey, err)
-		}
-		return k, nil
-	default:
-		return nil, fmt.Errorf(`invalid key type '%T' for jwk.New`, key)
-	}
+	return conv.Import(raw)
 }
 
 // PublicSetOf returns a new jwk.Set consisting of
@@ -134,7 +83,7 @@ func PublicSetOf(v Set) (Set, error) {
 	newSet := NewSet()
 
 	n := v.Len()
-	for i := 0; i < n; i++ {
+	for i := range n {
 		k, ok := v.Key(i)
 		if !ok {
 			return nil, fmt.Errorf(`key not found`)
@@ -160,15 +109,15 @@ func PublicSetOf(v Set) (Set, error) {
 // responsibility to remove any fields, if necessary
 //
 // If `v` is a raw key, the key is first converted to a `jwk.Key`
-func PublicKeyOf(v interface{}) (Key, error) {
+func PublicKeyOf(v any) (Key, error) {
 	// This should catch all jwk.Key instances
 	if pk, ok := v.(PublicKeyer); ok {
 		return pk.PublicKey()
 	}
 
-	jk, err := FromRaw(v)
+	jk, err := Import(v)
 	if err != nil {
-		return nil, fmt.Errorf(`failed to convert key into JWK: %w`, err)
+		return nil, fmt.Errorf(`jwk.PublicKeyOf: failed to convert key into JWK: %w`, err)
 	}
 
 	return jk.PublicKey()
@@ -181,198 +130,46 @@ func PublicKeyOf(v interface{}) (Key, error) {
 // The returned value will always be a pointer to the public key,
 // except when a []byte (e.g. symmetric key, ed25519 key) is passed to `v`.
 // In this case, the same []byte value is returned.
-func PublicRawKeyOf(v interface{}) (interface{}, error) {
-	if pk, ok := v.(PublicKeyer); ok {
-		pubk, err := pk.PublicKey()
-		if err != nil {
-			return nil, fmt.Errorf(`failed to obtain public key from %T: %w`, v, err)
-		}
-
-		var raw interface{}
-		if err := pubk.Raw(&raw); err != nil {
-			return nil, fmt.Errorf(`failed to obtain raw key from %T: %w`, pubk, err)
-		}
-		return raw, nil
-	}
-
-	// This may be a silly idea, but if the user gave us a non-pointer value...
-	var ptr interface{}
-	switch v := v.(type) {
-	case rsa.PrivateKey:
-		ptr = &v
-	case rsa.PublicKey:
-		ptr = &v
-	case ecdsa.PrivateKey:
-		ptr = &v
-	case ecdsa.PublicKey:
-		ptr = &v
-	default:
-		ptr = v
-	}
-
-	switch x := ptr.(type) {
-	case *rsa.PrivateKey:
-		return &x.PublicKey, nil
-	case *rsa.PublicKey:
-		return x, nil
-	case *ecdsa.PrivateKey:
-		return &x.PublicKey, nil
-	case *ecdsa.PublicKey:
-		return x, nil
-	case ed25519.PrivateKey:
-		return x.Public(), nil
-	case ed25519.PublicKey:
-		return x, nil
-	case x25519.PrivateKey:
-		return x.Public(), nil
-	case x25519.PublicKey:
-		return x, nil
-	case []byte:
-		return x, nil
-	default:
-		return nil, fmt.Errorf(`invalid key type passed to PublicKeyOf (%T)`, v)
-	}
-}
-
-const (
-	pmPrivateKey    = `PRIVATE KEY`
-	pmPublicKey     = `PUBLIC KEY`
-	pmECPrivateKey  = `EC PRIVATE KEY`
-	pmRSAPublicKey  = `RSA PUBLIC KEY`
-	pmRSAPrivateKey = `RSA PRIVATE KEY`
-)
-
-// EncodeX509 encodes the key into a byte sequence in ASN.1 DER format
-// suitable for to be PEM encoded. The key can be a jwk.Key or a raw key
-// instance, but it must be one of the types supported by `x509` package.
 //
-// This function will try to do the right thing depending on the key type
-// (i.e. switch between `x509.MarshalPKCS1PrivateKey` and `x509.MarshalECPrivateKey`),
-// but for public keys, it will always use `x509.MarshalPKIXPublicKey`.
-// Please manually perform the encoding if you need more fine-grained control
-//
-// The first return value is the name that can be used for `(pem.Block).Type`.
-// The second return value is the encoded byte sequence.
-func EncodeX509(v interface{}) (string, []byte, error) {
-	// we can't import jwk, so just use the interface
-	if key, ok := v.(interface{ Raw(interface{}) error }); ok {
-		var raw interface{}
-		if err := key.Raw(&raw); err != nil {
-			return "", nil, fmt.Errorf(`failed to get raw key out of %T: %w`, key, err)
+// This function must go through converting the object once to a jwk.Key,
+// then back to a raw key, so it's not exactly efficient.
+func PublicRawKeyOf(v any) (any, error) {
+	pk, ok := v.(PublicKeyer)
+	if !ok {
+		k, err := Import(v)
+		if err != nil {
+			return nil, fmt.Errorf(`jwk.PublicRawKeyOf: failed to convert key to jwk.Key: %w`, err)
 		}
 
-		v = raw
+		pk, ok = k.(PublicKeyer)
+		if !ok {
+			return nil, fmt.Errorf(`jwk.PublicRawKeyOf: failed to convert key to jwk.PublicKeyer: %w`, err)
+		}
 	}
 
-	// Try to convert it into a certificate
-	switch v := v.(type) {
-	case *rsa.PrivateKey:
-		return pmRSAPrivateKey, x509.MarshalPKCS1PrivateKey(v), nil
-	case *ecdsa.PrivateKey:
-		marshaled, err := x509.MarshalECPrivateKey(v)
-		if err != nil {
-			return "", nil, err
-		}
-		return pmECPrivateKey, marshaled, nil
-	case ed25519.PrivateKey:
-		marshaled, err := x509.MarshalPKCS8PrivateKey(v)
-		if err != nil {
-			return "", nil, err
-		}
-		return pmPrivateKey, marshaled, nil
-	case *rsa.PublicKey, *ecdsa.PublicKey, ed25519.PublicKey:
-		marshaled, err := x509.MarshalPKIXPublicKey(v)
-		if err != nil {
-			return "", nil, err
-		}
-		return pmPublicKey, marshaled, nil
-	default:
-		return "", nil, fmt.Errorf(`unsupported type %T for ASN.1 DER encoding`, v)
-	}
-}
-
-// EncodePEM encodes the key into a PEM encoded ASN.1 DER format.
-// The key can be a jwk.Key or a raw key instance, but it must be one of
-// the types supported by `x509` package.
-//
-// Internally, it uses the same routine as `jwk.EncodeX509()`, and therefore
-// the same caveats apply
-func EncodePEM(v interface{}) ([]byte, error) {
-	typ, marshaled, err := EncodeX509(v)
+	pubk, err := pk.PublicKey()
 	if err != nil {
-		return nil, fmt.Errorf(`failed to encode key in x509: %w`, err)
+		return nil, fmt.Errorf(`jwk.PublicRawKeyOf: failed to obtain public key from %T: %w`, v, err)
 	}
 
-	block := &pem.Block{
-		Type:  typ,
-		Bytes: marshaled,
-	}
-	return pem.EncodeToMemory(block), nil
-}
-
-// DecodePEM decodes a key in PEM encoded ASN.1 DER format.
-// and returns a raw key
-func DecodePEM(src []byte) (interface{}, []byte, error) {
-	block, rest := pem.Decode(src)
-	if block == nil {
-		return nil, nil, fmt.Errorf(`failed to decode PEM data`)
-	}
-
-	switch block.Type {
-	// Handle the semi-obvious cases
-	case pmRSAPrivateKey:
-		key, err := x509.ParsePKCS1PrivateKey(block.Bytes)
-		if err != nil {
-			return nil, nil, fmt.Errorf(`failed to parse PKCS1 private key: %w`, err)
-		}
-		return key, rest, nil
-	case pmRSAPublicKey:
-		key, err := x509.ParsePKCS1PublicKey(block.Bytes)
-		if err != nil {
-			return nil, nil, fmt.Errorf(`failed to parse PKCS1 public key: %w`, err)
-		}
-		return key, rest, nil
-	case pmECPrivateKey:
-		key, err := x509.ParseECPrivateKey(block.Bytes)
-		if err != nil {
-			return nil, nil, fmt.Errorf(`failed to parse EC private key: %w`, err)
-		}
-		return key, rest, nil
-	case pmPublicKey:
-		// XXX *could* return dsa.PublicKey
-		key, err := x509.ParsePKIXPublicKey(block.Bytes)
-		if err != nil {
-			return nil, nil, fmt.Errorf(`failed to parse PKIX public key: %w`, err)
-		}
-		return key, rest, nil
-	case pmPrivateKey:
-		key, err := x509.ParsePKCS8PrivateKey(block.Bytes)
-		if err != nil {
-			return nil, nil, fmt.Errorf(`failed to parse PKCS8 private key: %w`, err)
-		}
-		return key, rest, nil
-	case "CERTIFICATE":
-		cert, err := x509.ParseCertificate(block.Bytes)
-		if err != nil {
-			return nil, nil, fmt.Errorf(`failed to parse certificate: %w`, err)
-		}
-		return cert.PublicKey, rest, nil
-	default:
-		return nil, nil, fmt.Errorf(`invalid PEM block type %s`, block.Type)
+	var raw any
+	if err := Export(pubk, &raw); err != nil {
+		return nil, fmt.Errorf(`jwk.PublicRawKeyOf: failed to obtain raw key from %T: %w`, pubk, err)
 	}
+	return raw, nil
 }
 
 // ParseRawKey is a combination of ParseKey and Raw. It parses a single JWK key,
 // and assigns the "raw" key to the given parameter. The key must either be
 // a pointer to an empty interface, or a pointer to the actual raw key type
 // such as *rsa.PrivateKey, *ecdsa.PublicKey, *[]byte, etc.
-func ParseRawKey(data []byte, rawkey interface{}) error {
+func ParseRawKey(data []byte, rawkey any) error {
 	key, err := ParseKey(data)
 	if err != nil {
 		return fmt.Errorf(`failed to parse key: %w`, err)
 	}
 
-	if err := key.Raw(rawkey); err != nil {
+	if err := Export(key, rawkey); err != nil {
 		return fmt.Errorf(`failed to assign to raw key variable: %w`, err)
 	}
 
@@ -381,6 +178,7 @@ func ParseRawKey(data []byte, rawkey interface{}) error {
 
 type setDecodeCtx struct {
 	json.DecodeCtx
+
 	ignoreParseError bool
 }
 
@@ -402,18 +200,26 @@ func (ctx *setDecodeCtx) IgnoreParseError() bool {
 func ParseKey(data []byte, options ...ParseOption) (Key, error) {
 	var parsePEM bool
 	var localReg *json.Registry
+	var pemDecoder PEMDecoder
 	for _, option := range options {
-		//nolint:forcetypeassert
 		switch option.Ident() {
 		case identPEM{}:
-			parsePEM = option.Value().(bool)
+			if err := option.Value(&parsePEM); err != nil {
+				return nil, fmt.Errorf(`failed to retrieve PEM option value: %w`, err)
+			}
+		case identPEMDecoder{}:
+			if err := option.Value(&pemDecoder); err != nil {
+				return nil, fmt.Errorf(`failed to retrieve PEMDecoder option value: %w`, err)
+			}
 		case identLocalRegistry{}:
-			// in reality you can only pass either withLocalRegistry or
-			// WithTypedField, but since withLocalRegistry is used only by us,
-			// we skip checking
-			localReg = option.Value().(*json.Registry)
+			if err := option.Value(&localReg); err != nil {
+				return nil, fmt.Errorf(`failed to retrieve local registry option value: %w`, err)
+			}
 		case identTypedField{}:
-			pair := option.Value().(typedFieldPair)
+			var pair typedFieldPair // temporary var needed for typed field
+			if err := option.Value(&pair); err != nil {
+				return nil, fmt.Errorf(`failed to retrieve typed field option value: %w`, err)
+			}
 			if localReg == nil {
 				localReg = json.NewRegistry()
 			}
@@ -424,63 +230,49 @@ func ParseKey(data []byte, options ...ParseOption) (Key, error) {
 	}
 
 	if parsePEM {
-		raw, _, err := DecodePEM(data)
-		if err != nil {
-			return nil, fmt.Errorf(`failed to parse PEM encoded key: %w`, err)
-		}
-		return FromRaw(raw)
-	}
+		var raw any
 
-	var hint struct {
-		Kty string          `json:"kty"`
-		D   json.RawMessage `json:"d"`
-	}
-
-	if err := json.Unmarshal(data, &hint); err != nil {
-		return nil, fmt.Errorf(`failed to unmarshal JSON into key hint: %w`, err)
-	}
-
-	var key Key
-	switch jwa.KeyType(hint.Kty) {
-	case jwa.RSA:
-		if len(hint.D) > 0 {
-			key = newRSAPrivateKey()
+		// PEMDecoder should probably be deprecated, because of being a misnomer.
+		if pemDecoder != nil {
+			if err := decodeX509WithPEMDEcoder(&raw, data, pemDecoder); err != nil {
+				return nil, fmt.Errorf(`failed to decode PEM encoded key: %w`, err)
+			}
 		} else {
-			key = newRSAPublicKey()
+			// This version takes into account the various X509 decoders that are
+			// pre-registered.
+			if err := decodeX509(&raw, data); err != nil {
+				return nil, fmt.Errorf(`failed to decode X.509 encoded key: %w`, err)
+			}
 		}
-	case jwa.EC:
-		if len(hint.D) > 0 {
-			key = newECDSAPrivateKey()
-		} else {
-			key = newECDSAPublicKey()
-		}
-	case jwa.OctetSeq:
-		key = newSymmetricKey()
-	case jwa.OKP:
-		if len(hint.D) > 0 {
-			key = newOKPPrivateKey()
-		} else {
-			key = newOKPPublicKey()
-		}
-	default:
-		return nil, fmt.Errorf(`invalid key type from JSON (%s)`, hint.Kty)
+		return Import(raw)
 	}
 
-	if localReg != nil {
-		dcKey, ok := key.(json.DecodeCtxContainer)
-		if !ok {
-			return nil, fmt.Errorf(`typed field was requested, but the key (%T) does not support DecodeCtx`, key)
+	probe, err := keyProbe.Probe(data)
+	if err != nil {
+		return nil, fmt.Errorf(`jwk.Parse: failed to probe data: %w`, err)
+	}
+
+	unmarshaler := keyUnmarshaler{localReg: localReg}
+
+	muKeyParser.RLock()
+	parsers := make([]KeyParser, len(keyParsers))
+	copy(parsers, keyParsers)
+	muKeyParser.RUnlock()
+
+	for i := len(parsers) - 1; i >= 0; i-- {
+		parser := parsers[i]
+		key, err := parser.ParseKey(probe, &unmarshaler, data)
+		if err == nil {
+			return key, nil
 		}
-		dc := json.NewDecodeCtx(localReg)
-		dcKey.SetDecodeCtx(dc)
-		defer func() { dcKey.SetDecodeCtx(nil) }()
-	}
 
-	if err := json.Unmarshal(data, key); err != nil {
-		return nil, fmt.Errorf(`failed to unmarshal JSON into key (%T): %w`, key, err)
-	}
+		if errors.Is(err, ContinueError()) {
+			continue
+		}
 
-	return key, nil
+		return nil, err
+	}
+	return nil, fmt.Errorf(`jwk.Parse: no parser was able to parse the key`)
 }
 
 // Parse parses JWK from the incoming []byte.
@@ -499,17 +291,33 @@ func ParseKey(data []byte, options ...ParseOption) (Key, error) {
 // for `jwk.ParseKey()`.
 func Parse(src []byte, options ...ParseOption) (Set, error) {
 	var parsePEM bool
+	var parseX509 bool
 	var localReg *json.Registry
 	var ignoreParseError bool
+	var pemDecoder PEMDecoder
 	for _, option := range options {
-		//nolint:forcetypeassert
 		switch option.Ident() {
 		case identPEM{}:
-			parsePEM = option.Value().(bool)
+			if err := option.Value(&parsePEM); err != nil {
+				return nil, parseerr(`failed to retrieve PEM option value: %w`, err)
+			}
+		case identX509{}:
+			if err := option.Value(&parseX509); err != nil {
+				return nil, parseerr(`failed to retrieve X509 option value: %w`, err)
+			}
+		case identPEMDecoder{}:
+			if err := option.Value(&pemDecoder); err != nil {
+				return nil, parseerr(`failed to retrieve PEMDecoder option value: %w`, err)
+			}
 		case identIgnoreParseError{}:
-			ignoreParseError = option.Value().(bool)
+			if err := option.Value(&ignoreParseError); err != nil {
+				return nil, parseerr(`failed to retrieve IgnoreParseError option value: %w`, err)
+			}
 		case identTypedField{}:
-			pair := option.Value().(typedFieldPair)
+			var pair typedFieldPair // temporary var needed for typed field
+			if err := option.Value(&pair); err != nil {
+				return nil, parseerr(`failed to retrieve typed field option value: %w`, err)
+			}
 			if localReg == nil {
 				localReg = json.NewRegistry()
 			}
@@ -519,19 +327,22 @@ func Parse(src []byte, options ...ParseOption) (Set, error) {
 
 	s := NewSet()
 
-	if parsePEM {
+	if parsePEM || parseX509 {
+		if pemDecoder == nil {
+			pemDecoder = NewPEMDecoder()
+		}
 		src = bytes.TrimSpace(src)
 		for len(src) > 0 {
-			raw, rest, err := DecodePEM(src)
+			raw, rest, err := pemDecoder.Decode(src)
 			if err != nil {
-				return nil, fmt.Errorf(`failed to parse PEM encoded key: %w`, err)
+				return nil, parseerr(`failed to parse PEM encoded key: %w`, err)
 			}
-			key, err := FromRaw(raw)
+			key, err := Import(raw)
 			if err != nil {
-				return nil, fmt.Errorf(`failed to create jwk.Key from %T: %w`, raw, err)
+				return nil, parseerr(`failed to create jwk.Key from %T: %w`, raw, err)
 			}
 			if err := s.AddKey(key); err != nil {
-				return nil, fmt.Errorf(`failed to add jwk.Key to set: %w`, err)
+				return nil, parseerr(`failed to add jwk.Key to set: %w`, err)
 			}
 			src = bytes.TrimSpace(rest)
 		}
@@ -541,7 +352,7 @@ func Parse(src []byte, options ...ParseOption) (Set, error) {
 	if localReg != nil || ignoreParseError {
 		dcKs, ok := s.(KeyWithDecodeCtx)
 		if !ok {
-			return nil, fmt.Errorf(`typed field was requested, but the key set (%T) does not support DecodeCtx`, s)
+			return nil, parseerr(`typed field was requested, but the key set (%T) does not support DecodeCtx`, s)
 		}
 		dc := &setDecodeCtx{
 			DecodeCtx:        json.NewDecodeCtx(localReg),
@@ -552,7 +363,7 @@ func Parse(src []byte, options ...ParseOption) (Set, error) {
 	}
 
 	if err := json.Unmarshal(src, s); err != nil {
-		return nil, fmt.Errorf(`failed to unmarshal JWK set: %w`, err)
+		return nil, parseerr(`failed to unmarshal JWK set: %w`, err)
 	}
 
 	return s, nil
@@ -564,31 +375,40 @@ func ParseReader(src io.Reader, options ...ParseOption) (Set, error) {
 	// JWKs except when we encounter an EOF, so just... ReadAll
 	buf, err := io.ReadAll(src)
 	if err != nil {
-		return nil, fmt.Errorf(`failed to read from io.Reader: %w`, err)
+		return nil, rparseerr(`failed to read from io.Reader: %w`, err)
 	}
 
-	return Parse(buf, options...)
+	set, err := Parse(buf, options...)
+	if err != nil {
+		return nil, rparseerr(`failed to parse reader: %w`, err)
+	}
+	return set, nil
 }
 
 // ParseString parses a JWK set from the incoming string.
 func ParseString(s string, options ...ParseOption) (Set, error) {
-	return Parse([]byte(s), options...)
+	set, err := Parse([]byte(s), options...)
+	if err != nil {
+		return nil, sparseerr(`failed to parse string: %w`, err)
+	}
+	return set, nil
 }
 
 // AssignKeyID is a convenience function to automatically assign the "kid"
 // section of the key, if it already doesn't have one. It uses Key.Thumbprint
 // method with crypto.SHA256 as the default hashing algorithm
 func AssignKeyID(key Key, options ...AssignKeyIDOption) error {
-	if _, ok := key.Get(KeyIDKey); ok {
+	if key.Has(KeyIDKey) {
 		return nil
 	}
 
 	hash := crypto.SHA256
 	for _, option := range options {
-		//nolint:forcetypeassert
 		switch option.Ident() {
 		case identThumbprintHash{}:
-			hash = option.Value().(crypto.Hash)
+			if err := option.Value(&hash); err != nil {
+				return fmt.Errorf(`failed to retrieve thumbprint hash option value: %w`, err)
+			}
 		}
 	}
 
@@ -604,6 +424,7 @@ func AssignKeyID(key Key, options ...AssignKeyIDOption) error {
 	return nil
 }
 
+// NOTE: may need to remove this to allow pluggale key types
 func cloneKey(src Key) (Key, error) {
 	var dst Key
 	switch src.(type) {
@@ -622,14 +443,17 @@ func cloneKey(src Key) (Key, error) {
 	case SymmetricKey:
 		dst = newSymmetricKey()
 	default:
-		return nil, fmt.Errorf(`unknown key type %T`, src)
+		return nil, fmt.Errorf(`jwk.cloneKey: unknown key type %T`, src)
 	}
 
-	for _, pair := range src.makePairs() {
-		//nolint:forcetypeassert
-		key := pair.Key.(string)
-		if err := dst.Set(key, pair.Value); err != nil {
-			return nil, fmt.Errorf(`failed to set %q: %w`, key, err)
+	for _, k := range src.Keys() {
+		// It's absolutely
+		var v any
+		if err := src.Get(k, &v); err != nil {
+			return nil, fmt.Errorf(`jwk.cloneKey: failed to get %q: %w`, k, err)
+		}
+		if err := dst.Set(k, v); err != nil {
+			return nil, fmt.Errorf(`jwk.cloneKey: failed to set %q: %w`, k, err)
 		}
 	}
 	return dst, nil
@@ -643,7 +467,7 @@ func cloneKey(src Key) (Key, error) {
 //
 // Currently only EC (including Ed25519) and RSA keys (and jwk.Set
 // comprised of these key types) are supported.
-func Pem(v interface{}) ([]byte, error) {
+func Pem(v any) ([]byte, error) {
 	var set Set
 	switch v := v.(type) {
 	case Key:
@@ -658,7 +482,7 @@ func Pem(v interface{}) ([]byte, error) {
 	}
 
 	var ret []byte
-	for i := 0; i < set.Len(); i++ {
+	for i := range set.Len() {
 		key, _ := set.Key(i)
 		typ, buf, err := asnEncode(key)
 		if err != nil {
@@ -677,7 +501,7 @@ func asnEncode(key Key) (string, []byte, error) {
 	switch key := key.(type) {
 	case ECDSAPrivateKey:
 		var rawkey ecdsa.PrivateKey
-		if err := key.Raw(&rawkey); err != nil {
+		if err := Export(key, &rawkey); err != nil {
 			return "", nil, fmt.Errorf(`failed to get raw key from jwk.Key: %w`, err)
 		}
 		buf, err := x509.MarshalECPrivateKey(&rawkey)
@@ -686,8 +510,8 @@ func asnEncode(key Key) (string, []byte, error) {
 		}
 		return pmECPrivateKey, buf, nil
 	case RSAPrivateKey, OKPPrivateKey:
-		var rawkey interface{}
-		if err := key.Raw(&rawkey); err != nil {
+		var rawkey any
+		if err := Export(key, &rawkey); err != nil {
 			return "", nil, fmt.Errorf(`failed to get raw key from jwk.Key: %w`, err)
 		}
 		buf, err := x509.MarshalPKCS8PrivateKey(rawkey)
@@ -696,8 +520,8 @@ func asnEncode(key Key) (string, []byte, error) {
 		}
 		return pmPrivateKey, buf, nil
 	case RSAPublicKey, ECDSAPublicKey, OKPPublicKey:
-		var rawkey interface{}
-		if err := key.Raw(&rawkey); err != nil {
+		var rawkey any
+		if err := Export(key, &rawkey); err != nil {
 			return "", nil, fmt.Errorf(`failed to get raw key from jwk.Key: %w`, err)
 		}
 		buf, err := x509.MarshalPKIXPublicKey(rawkey)
@@ -710,6 +534,9 @@ func asnEncode(key Key) (string, []byte, error) {
 	}
 }
 
+type CustomDecoder = json.CustomDecoder
+type CustomDecodeFunc = json.CustomDecodeFunc
+
 // RegisterCustomField allows users to specify that a private field
 // be decoded as an instance of the specified type. This option has
 // a global effect.
@@ -718,27 +545,38 @@ func asnEncode(key Key) (string, []byte, error) {
 // you want to represent as a string formatted in RFC3339 in JSON,
 // but want it back as `time.Time`.
 //
-// In that case you would register a custom field as follows
+// In such case you would register a custom field as follows
 //
-//	jwk.RegisterCustomField(`x-birthday`, timeT)
+//	jwk.RegisterCustomField(`x-birthday`, time.Time{})
 //
-// Then `key.Get("x-birthday")` will still return an `interface{}`,
-// but you can convert its type to `time.Time`
+// Then you can use a `time.Time` variable to extract the value
+// of `x-birthday` field, instead of having to use `any`
+// and later convert it to `time.Time`
 //
-//	bdayif, _ := key.Get(`x-birthday`)
-//	bday := bdayif.(time.Time)
-func RegisterCustomField(name string, object interface{}) {
+//	var bday time.Time
+//	_ = key.Get(`x-birthday`, &bday)
+//
+// If you need a more fine-tuned control over the decoding process,
+// you can register a `CustomDecoder`. For example, below shows
+// how to register a decoder that can parse RFC1123 format string:
+//
+//	jwk.RegisterCustomField(`x-birthday`, jwk.CustomDecodeFunc(func(data []byte) (any, error) {
+//	  return time.Parse(time.RFC1123, string(data))
+//	}))
+//
+// Please note that use of custom fields can be problematic if you
+// are using a library that does not implement MarshalJSON/UnmarshalJSON
+// and you try to roundtrip from an object to JSON, and then back to an object.
+// For example, in the above example, you can _parse_ time values formatted
+// in the format specified in RFC822, but when you convert an object into
+// JSON, it will be formatted in RFC3339, because that's what `time.Time`
+// likes to do. To avoid this, it's always better to use a custom type
+// that wraps your desired type (in this case `time.Time`) and implement
+// MarshalJSON and UnmashalJSON.
+func RegisterCustomField(name string, object any) {
 	registry.Register(name, object)
 }
 
-func AvailableCurves() []elliptic.Curve {
-	return ecutil.AvailableCurves()
-}
-
-func CurveForAlgorithm(alg jwa.EllipticCurveAlgorithm) (elliptic.Curve, bool) {
-	return ecutil.CurveForAlgorithm(alg)
-}
-
 // Equal compares two keys and returns true if they are equal. The comparison
 // is solely done against the thumbprints of k1 and k2. It is possible for keys
 // that have, for example, different key IDs, key usage, etc, to be considered equal.
@@ -797,3 +635,76 @@ func IsKeyValidationError(err error) bool {
 	var kve keyValidationError
 	return errors.Is(err, &kve)
 }
+
+// Configure is used to configure global behavior of the jwk package.
+func Configure(options ...GlobalOption) {
+	var strictKeyUsagePtr *bool
+	for _, option := range options {
+		switch option.Ident() {
+		case identStrictKeyUsage{}:
+			var v bool
+			if err := option.Value(&v); err != nil {
+				continue
+			}
+			strictKeyUsagePtr = &v
+		}
+	}
+
+	if strictKeyUsagePtr != nil {
+		strictKeyUsage.Store(*strictKeyUsagePtr)
+	}
+}
+
+// These are used when validating keys.
+type keyWithD interface {
+	D() ([]byte, bool)
+}
+
+var _ keyWithD = &okpPrivateKey{}
+
+func extractEmbeddedKey(keyif Key, concretTypes []reflect.Type) (Key, error) {
+	rv := reflect.ValueOf(keyif)
+
+	// If the value can be converted to one of the concrete types, then we're done
+	if slices.ContainsFunc(concretTypes, func(t reflect.Type) bool {
+		return rv.Type().ConvertibleTo(t)
+	}) {
+		return keyif, nil
+	}
+
+	// When a struct implements the Key interface via embedding, you unfortunately
+	// cannot use a type switch to determine the concrete type, because
+	if rv.Kind() == reflect.Ptr {
+		if rv.IsNil() {
+			return nil, fmt.Errorf(`invalid key value (0): %w`, ContinueError())
+		}
+		rv = rv.Elem()
+	}
+
+	if rv.Kind() != reflect.Struct {
+		return nil, fmt.Errorf(`invalid key value type %T (1): %w`, keyif, ContinueError())
+	}
+	if rv.NumField() == 0 {
+		return nil, fmt.Errorf(`invalid key value type %T (2): %w`, keyif, ContinueError())
+	}
+	// Iterate through the fields of the struct to find the first field that
+	// implements the Key interface
+	rt := rv.Type()
+	for i := range rv.NumField() {
+		field := rv.Field(i)
+		ft := rt.Field(i)
+		if !ft.Anonymous {
+			// We can only salvage this object if the object implements jwk.Key
+			// via embedding, so we skip fields that are not anonymous
+			continue
+		}
+
+		if field.CanInterface() {
+			if k, ok := field.Interface().(Key); ok {
+				return extractEmbeddedKey(k, concretTypes)
+			}
+		}
+	}
+
+	return nil, fmt.Errorf(`invalid key value type %T (3): %w`, keyif, ContinueError())
+}
