@@ -6,52 +6,56 @@ package jwt
 
 import (
 	"bytes"
-	"errors"
 	"fmt"
 	"io"
 	"sync/atomic"
+	"time"
 
-	"github.com/lestrrat-go/jwx/v2"
-	"github.com/lestrrat-go/jwx/v2/internal/json"
-	"github.com/lestrrat-go/jwx/v2/jws"
-	"github.com/lestrrat-go/jwx/v2/jwt/internal/types"
+	"github.com/lestrrat-go/jwx/v3"
+	"github.com/lestrrat-go/jwx/v3/internal/json"
+	"github.com/lestrrat-go/jwx/v3/jwa"
+	"github.com/lestrrat-go/jwx/v3/jws"
+	jwterrs "github.com/lestrrat-go/jwx/v3/jwt/internal/errors"
+	"github.com/lestrrat-go/jwx/v3/jwt/internal/types"
 )
 
-var compactOnly uint32
-var errInvalidJWT = errors.New(`invalid JWT`)
-
-// ErrInvalidJWT returns the opaque error value that is returned when
-// `jwt.Parse` fails due to not being able to deduce the format of
-// the incoming buffer
-func ErrInvalidJWT() error {
-	return errInvalidJWT
-}
+var defaultTruncation atomic.Int64
 
 // Settings controls global settings that are specific to JWTs.
 func Settings(options ...GlobalOption) {
 	var flattenAudience bool
-	var compactOnlyBool bool
 	var parsePedantic bool
 	var parsePrecision = types.MaxPrecision + 1  // illegal value, so we can detect nothing was set
 	var formatPrecision = types.MaxPrecision + 1 // illegal value, so we can detect nothing was set
-
-	//nolint:forcetypeassert
+	truncation := time.Duration(-1)
 	for _, option := range options {
 		switch option.Ident() {
+		case identTruncation{}:
+			if err := option.Value(&truncation); err != nil {
+				panic(fmt.Sprintf("jwt.Settings: value for WithTruncation must be time.Duration: %s", err))
+			}
 		case identFlattenAudience{}:
-			flattenAudience = option.Value().(bool)
-		case identCompactOnly{}:
-			compactOnlyBool = option.Value().(bool)
+			if err := option.Value(&flattenAudience); err != nil {
+				panic(fmt.Sprintf("jwt.Settings: value for WithFlattenAudience must be bool: %s", err))
+			}
 		case identNumericDateParsePedantic{}:
-			parsePedantic = option.Value().(bool)
+			if err := option.Value(&parsePedantic); err != nil {
+				panic(fmt.Sprintf("jwt.Settings: value for WithNumericDateParsePedantic must be bool: %s", err))
+			}
 		case identNumericDateParsePrecision{}:
-			v := option.Value().(int)
+			var v int
+			if err := option.Value(&v); err != nil {
+				panic(fmt.Sprintf("jwt.Settings: value for WithNumericDateParsePrecision must be int: %s", err))
+			}
 			// only accept this value if it's in our desired range
 			if v >= 0 && v <= int(types.MaxPrecision) {
 				parsePrecision = uint32(v)
 			}
 		case identNumericDateFormatPrecision{}:
-			v := option.Value().(int)
+			var v int
+			if err := option.Value(&v); err != nil {
+				panic(fmt.Sprintf("jwt.Settings: value for WithNumericDateFormatPrecision must be int: %s", err))
+			}
 			// only accept this value if it's in our desired range
 			if v >= 0 && v <= int(types.MaxPrecision) {
 				formatPrecision = uint32(v)
@@ -84,17 +88,6 @@ func Settings(options ...GlobalOption) {
 		}
 	}
 
-	{
-		v := atomic.LoadUint32(&compactOnly)
-		if (v == 1) != compactOnlyBool {
-			var newVal uint32
-			if compactOnlyBool {
-				newVal = 1
-			}
-			atomic.CompareAndSwapUint32(&compactOnly, v, newVal)
-		}
-	}
-
 	{
 		defaultOptionsMu.Lock()
 		if flattenAudience {
@@ -104,20 +97,28 @@ func Settings(options ...GlobalOption) {
 		}
 		defaultOptionsMu.Unlock()
 	}
+
+	if truncation >= 0 {
+		defaultTruncation.Store(int64(truncation))
+	}
 }
 
 var registry = json.NewRegistry()
 
 // ParseString calls Parse against a string
 func ParseString(s string, options ...ParseOption) (Token, error) {
-	return parseBytes([]byte(s), options...)
+	tok, err := parseBytes([]byte(s), options...)
+	if err != nil {
+		return nil, jwterrs.ParseErrorf(`jwt.ParseString`, `failed to parse string: %w`, err)
+	}
+	return tok, nil
 }
 
 // Parse parses the JWT token payload and creates a new `jwt.Token` object.
-// The token must be encoded in either JSON format or compact format.
+// The token must be encoded in JWS compact format, or a raw JSON form of JWT
+// without any signatures.
 //
-// This function can only work with either raw JWT (JSON) and JWS (Compact or JSON).
-// If you need JWE support on top of it, you will need to rollout your
+// If you need JWE support on top of JWS, you will need to rollout your
 // own workaround.
 //
 // If the token is signed, and you want to verify the payload matches the signature,
@@ -139,7 +140,11 @@ func ParseString(s string, options ...ParseOption) (Token, error) {
 // ParseOptions control the parsing behavior, and ValidateOptions are
 // passed to `Validate()` when `jwt.WithValidate` is specified.
 func Parse(s []byte, options ...ParseOption) (Token, error) {
-	return parseBytes(s, options...)
+	tok, err := parseBytes(s, options...)
+	if err != nil {
+		return nil, jwterrs.ParseErrorf(`jwt.Parse`, `failed to parse token: %w`, err)
+	}
+	return tok, nil
 }
 
 // ParseInsecure is exactly the same as Parse(), but it disables
@@ -152,12 +157,16 @@ func ParseInsecure(s []byte, options ...ParseOption) (Token, error) {
 	for _, option := range options {
 		switch option.Ident() {
 		case identVerify{}, identValidate{}:
-			return nil, fmt.Errorf(`jwt.ParseInsecure: jwt.WithVerify() and jwt.WithValidate() may not be specified`)
+			return nil, jwterrs.ParseErrorf(`jwt.ParseInsecure`, `jwt.WithVerify() and jwt.WithValidate() may not be specified`)
 		}
 	}
 
 	options = append(options, WithVerify(false), WithValidate(false))
-	return Parse(s, options...)
+	tok, err := Parse(s, options...)
+	if err != nil {
+		return nil, jwterrs.ParseErrorf(`jwt.ParseInsecure`, `failed to parse token: %w`, err)
+	}
+	return tok, nil
 }
 
 // ParseReader calls Parse against an io.Reader
@@ -165,9 +174,13 @@ func ParseReader(src io.Reader, options ...ParseOption) (Token, error) {
 	// We're going to need the raw bytes regardless. Read it.
 	data, err := io.ReadAll(src)
 	if err != nil {
-		return nil, fmt.Errorf(`failed to read from token data source: %w`, err)
+		return nil, jwterrs.ParseErrorf(`jwt.ParseReader`, `failed to read from token data source: %w`, err)
 	}
-	return parseBytes(data, options...)
+	tok, err := parseBytes(data, options...)
+	if err != nil {
+		return nil, jwterrs.ParseErrorf(`jwt.ParseReader`, `failed to parse token: %w`, err)
+	}
+	return tok, nil
 }
 
 type parseCtx struct {
@@ -178,6 +191,8 @@ type parseCtx struct {
 	pedantic         bool
 	skipVerification bool
 	validate         bool
+	withKeyCount     int
+	withKey          *withKey // this is used to detect if we have a WithKey option
 }
 
 func parseBytes(data []byte, options ...ParseOption) (Token, error) {
@@ -196,27 +211,53 @@ func parseBytes(data []byte, options ...ParseOption) (Token, error) {
 	for _, o := range options {
 		if v, ok := o.(ValidateOption); ok {
 			ctx.validateOpts = append(ctx.validateOpts, v)
-			continue
+			// context is used for both verification and validation, so we can't just continue
+			switch o.Ident() {
+			case identContext{}:
+			default:
+				continue
+			}
 		}
 
-		//nolint:forcetypeassert
 		switch o.Ident() {
-		case identKey{}, identKeySet{}, identVerifyAuto{}, identKeyProvider{}:
+		case identKey{}:
+			// it would be nice to be able to detect if ctx.verifyOpts[0]
+			// is a WithKey option, but unfortunately at that point we have
+			// already converted the options to a jws option, which means
+			// we can no longer compare its Ident() to jwt.identKey{}.
+			// So let's just count this here
+			ctx.withKeyCount++
+			if ctx.withKeyCount == 1 {
+				if err := o.Value(&ctx.withKey); err != nil {
+					return nil, fmt.Errorf("jws.parseBytes: value for WithKey option must be a *jwt.withKey: %w", err)
+				}
+			}
+			verifyOpts = append(verifyOpts, o)
+		case identKeySet{}, identVerifyAuto{}, identKeyProvider{}, identBase64Encoder{}, identContext{}:
 			verifyOpts = append(verifyOpts, o)
 		case identToken{}:
-			token, ok := o.Value().(Token)
-			if !ok {
-				return nil, fmt.Errorf(`invalid token passed via WithToken() option (%T)`, o.Value())
+			var token Token
+			if err := o.Value(&token); err != nil {
+				return nil, fmt.Errorf("jws.parseBytes: value for WithToken option must be a jwt.Token: %w", err)
 			}
 			ctx.token = token
 		case identPedantic{}:
-			ctx.pedantic = o.Value().(bool)
+			if err := o.Value(&ctx.pedantic); err != nil {
+				return nil, fmt.Errorf("jws.parseBytes: value for WithPedantic option must be a bool: %w", err)
+			}
 		case identValidate{}:
-			ctx.validate = o.Value().(bool)
+			if err := o.Value(&ctx.validate); err != nil {
+				return nil, fmt.Errorf("jws.parseBytes: value for WithValidate option must be a bool: %w", err)
+			}
 		case identVerify{}:
-			verification = o.Value().(bool)
+			if err := o.Value(&verification); err != nil {
+				return nil, fmt.Errorf("jws.parseBytes: value for WithVerify option must be a bool: %w", err)
+			}
 		case identTypedClaim{}:
-			pair := o.Value().(claimPair)
+			var pair claimPair
+			if err := o.Value(&pair); err != nil {
+				return nil, fmt.Errorf("jws.parseBytes: value for WithTypedClaim option must be claimPair: %w", err)
+			}
 			if ctx.localReg == nil {
 				ctx.localReg = json.NewRegistry()
 			}
@@ -255,14 +296,24 @@ const (
 var _ = _JwsVerifyInvalid
 
 func verifyJWS(ctx *parseCtx, payload []byte) ([]byte, int, error) {
-	if len(ctx.verifyOpts) == 0 {
+	lvo := len(ctx.verifyOpts)
+	if lvo == 0 {
 		return nil, _JwsVerifySkipped, nil
 	}
 
-	verifyOpts := ctx.verifyOpts
-	if atomic.LoadUint32(&compactOnly) == 1 {
-		verifyOpts = append(verifyOpts, jws.WithCompact())
+	if lvo == 1 && ctx.withKeyCount == 1 {
+		wk := ctx.withKey
+		alg, ok := wk.alg.(jwa.SignatureAlgorithm)
+		if ok && len(wk.options) == 0 {
+			verified, err := jws.VerifyCompactFast(wk.key, payload, alg)
+			if err != nil {
+				return nil, _JwsVerifyDone, err
+			}
+			return verified, _JwsVerifyDone, nil
+		}
 	}
+
+	verifyOpts := append(ctx.verifyOpts, jws.WithCompact())
 	verified, err := jws.Verify(payload, verifyOpts...)
 	return verified, _JwsVerifyDone, err
 }
@@ -277,7 +328,7 @@ func parse(ctx *parseCtx, data []byte) (Token, error) {
 	var expectNested bool
 
 OUTER:
-	for i := 0; i < maxDecodeLevels; i++ {
+	for i := range maxDecodeLevels {
 		switch kind := jwx.GuessFormat(payload); kind {
 		case jwx.JWT:
 			if ctx.pedantic {
@@ -297,7 +348,7 @@ OUTER:
 
 			break OUTER
 		case jwx.InvalidFormat:
-			return nil, ErrInvalidJWT()
+			return nil, UnknownPayloadTypeError()
 		case jwx.UnknownFormat:
 			// "Unknown" may include invalid JWTs, for example, those who lack "aud"
 			// claim. We could be pedantic and reject these
@@ -349,11 +400,7 @@ OUTER:
 			}
 
 			// No verification.
-			var parseOptions []jws.ParseOption
-			if atomic.LoadUint32(&compactOnly) == 1 {
-				parseOptions = append(parseOptions, jws.WithCompact())
-			}
-			m, err := jws.Parse(data, parseOptions...)
+			m, err := jws.Parse(data, jws.WithCompact())
 			if err != nil {
 				return nil, fmt.Errorf(`invalid jws message: %w`, err)
 			}
@@ -397,6 +444,11 @@ OUTER:
 // or a jwk.Key, and the name of the algorithm that should be used to sign
 // the token.
 //
+// For well-known algorithms with no special considerations (e.g. detached
+// payloads, extra protected heders, etc), this function will automatically
+// take the fast path and bypass the jws.Sign() machinery, which improves
+// performance significantly.
+//
 // If the key is a jwk.Key and the key contains a key ID (`kid` field),
 // then it is added to the protected header generated by the signature
 //
@@ -411,6 +463,26 @@ OUTER:
 // to the literal value `JWT`, unless you provide a custom value for it
 // by jws.WithProtectedHeaders option, that can be passed to `jwt.WithKey“.
 func Sign(t Token, options ...SignOption) ([]byte, error) {
+	// fast path; can only happen if there is exactly one option
+	if len(options) == 1 && (options[0].Ident() == identKey{}) {
+		// The option must be a withKey option.
+		var wk *withKey
+		if err := options[0].Value(&wk); err == nil {
+			alg, ok := wk.alg.(jwa.SignatureAlgorithm)
+			if !ok {
+				return nil, fmt.Errorf(`jwt.Sign: invalid algorithm type %T. jwa.SignatureAlgorithm is required`, wk.alg)
+			}
+
+			// Check if option contains anything other than alg/key
+			if len(wk.options) == 0 {
+				// yay, we have something we can put in the FAST PATH!
+				return signFast(t, alg, wk.key)
+			}
+			// fallthrough
+		}
+		// fallthrough
+	}
+
 	var soptions []jws.SignOption
 	if l := len(options); l > 0 {
 		// we need to from SignOption to Option because ... reasons
@@ -465,16 +537,21 @@ func (t *stdToken) Clone() (Token, error) {
 	dst := New()
 
 	dst.Options().Set(*(t.Options()))
-	for _, pair := range t.makePairs() {
-		//nolint:forcetypeassert
-		key := pair.Key.(string)
-		if err := dst.Set(key, pair.Value); err != nil {
-			return nil, fmt.Errorf(`failed to set %s: %w`, key, err)
+	for _, k := range t.Keys() {
+		var v any
+		if err := t.Get(k, &v); err != nil {
+			return nil, fmt.Errorf(`jwt.Clone: failed to get %s: %w`, k, err)
+		}
+		if err := dst.Set(k, v); err != nil {
+			return nil, fmt.Errorf(`jwt.Clone failed to set %s: %w`, k, err)
 		}
 	}
 	return dst, nil
 }
 
+type CustomDecoder = json.CustomDecoder
+type CustomDecodeFunc = json.CustomDecodeFunc
+
 // RegisterCustomField allows users to specify that a private field
 // be decoded as an instance of the specified type. This option has
 // a global effect.
@@ -483,15 +560,38 @@ func (t *stdToken) Clone() (Token, error) {
 // you want to represent as a string formatted in RFC3339 in JSON,
 // but want it back as `time.Time`.
 //
-// In that case you would register a custom field as follows
+// In such case you would register a custom field as follows
 //
-//	jwt.RegisterCustomField(`x-birthday`, timeT)
+//	jwt.RegisterCustomField(`x-birthday`, time.Time{})
 //
-// Then `token.Get("x-birthday")` will still return an `interface{}`,
-// but you can convert its type to `time.Time`
+// Then you can use a `time.Time` variable to extract the value
+// of `x-birthday` field, instead of having to use `any`
+// and later convert it to `time.Time`
 //
-//	bdayif, _ := token.Get(`x-birthday`)
-//	bday := bdayif.(time.Time)
-func RegisterCustomField(name string, object interface{}) {
+//	var bday time.Time
+//	_ = token.Get(`x-birthday`, &bday)
+//
+// If you need a more fine-tuned control over the decoding process,
+// you can register a `CustomDecoder`. For example, below shows
+// how to register a decoder that can parse RFC822 format string:
+//
+//	jwt.RegisterCustomField(`x-birthday`, jwt.CustomDecodeFunc(func(data []byte) (any, error) {
+//	  return time.Parse(time.RFC822, string(data))
+//	}))
+//
+// Please note that use of custom fields can be problematic if you
+// are using a library that does not implement MarshalJSON/UnmarshalJSON
+// and you try to roundtrip from an object to JSON, and then back to an object.
+// For example, in the above example, you can _parse_ time values formatted
+// in the format specified in RFC822, but when you convert an object into
+// JSON, it will be formatted in RFC3339, because that's what `time.Time`
+// likes to do. To avoid this, it's always better to use a custom type
+// that wraps your desired type (in this case `time.Time`) and implement
+// MarshalJSON and UnmashalJSON.
+func RegisterCustomField(name string, object any) {
 	registry.Register(name, object)
 }
+
+func getDefaultTruncation() time.Duration {
+	return time.Duration(defaultTruncation.Load())
+}
